from orgasm.graph import DiGraph
import sys
from Carbon.Files import bNoDeskItems
from psycopg2.errorcodes import OUT_OF_MEMORY

def findDeepestRead(index,seed):
    frg   = [seed[x:(x+index.getReadSize())] for x in xrange(len(seed)-index.getReadSize())]
    fcount= [index.count(x) for x in frg]
    mfc = max(fcount)
    return frg[fcount.index(mfc)],mfc


class Assembler:
    
    def __init__(self,index,seed,overlap=90):
        '''
        Create a new assembler object.
        
        :param index: the file name of an read index previously formated using orgasmi binary
        :type index: bytes
        :param seed: a DNA sequenced used as started for the assembly
        :type seed: bytes
        :param overlap: the minimum overlap between two reads
        :type overlap: int
        '''
        starter,depth = findDeepestRead(index,seed)
        
        self._graph = DiGraph("assembling")
        
        self._index = index
        self._overlap=overlap
        
        reads = self.readType([x for x in index[starter]])
        
        self._graph.addNode(reads[0])
        attr = self._graph[reads[0]]
        attr["weight"]=reads[1]
        attr["reads"]=reads[2]
        
        self.extensionPoints = [(starter,reads[0])]
        
    def readType(self,ids):
        '''
        Internal function : Given a set of read ids, return the one that have to be used
        as standard id. 
        
        The set of ids given to this function corresponds to a set of all strictly identical reads.
         
        :param ids: an iterable elements contining read ids.
        :type ids: iterable
        
        :returns: a tuble of three elements
        
                    - the standard id to use
                    - the length of ids
                    - a set containing all unique ids given as parameter
                    
        :rtype: tuple
        '''
        if not ids:
            return None
        
        t = len(self._index)+1
        
        for y in ids:
            if abs(y) < abs(t):
                t = y
                
        return (t,len(ids),set(ids))
    
    def run(self):
        cycle = 0
        extcount = len(self.extensionPoints)
        
        while self.extensionPoints and extcount < 1000 :
            cycle+=1
            extcount = len(self.extensionPoints)
            print >>sys.stderr,"Cycle : %8d     Extension point : %4d \r" % (cycle,extcount),
            self.runOneCycle()
    
    def runOneCycle(self):
        extensions = []
        
        for e in self.extensionPoints:
            extensions.extend(self.nextExtention(*e))
            
        
        self.extensionPoints=set(extensions)
        
    def nextExtention(self,seed,nodeindex):
        i = self._index.getReadSize() - 1;
        
        e = {}
        
        while i >=self._overlap and not e:
            s = seed[-i:]
            e = self._index.getExtensions(s)
            i-=1
            
        starter = []
        for ex in e:
            data = e[ex]
            reads = self.readType(data[1])
            nodeindexE = reads[0]
            
            self._graph.addEdge(nodeindex, nodeindexE)
            
            attr = self._graph[nodeindexE]
            attr["weight"]=reads[1]
            attr["reads"]=reads[2]
            
            attr = self._graph[(nodeindex, nodeindexE)]
            attr["label"]=ex
            attr["arrowsize"]=float(reads[1]/10)
            attr["type"]="next read"
            
            
            son = set(self._graph.neighbourIterator(nodeindexE,edgePredicate=lambda e : self._graph[e]['type']=="next read"))
            if not son:
                starter.append((self._index.getRead(nodeindexE,0,self._index.getReadSize()),nodeindexE))            
            
        return starter
    
    def forkIterator(self):
        
        def has_several_sons(node):
            return  len(set(self._graph.neighbourIterator(node,edgePredicate=lambda e : self._graph[e]['type']=="next read"))) > 1
        
        return self._graph.nodeIterator(has_several_sons)
    

        
        
        

    def microBubble(self,startnode,lenmax=None, wide=100):
    
        def sons(node):
            return set(self._graph.neighbourIterator(node,edgePredicate=lambda e : self._graph[e]['type']=="next read"))
        
        def intersect(runs):
            return reduce(lambda x,y: x & y,(set(r[3]) for r in runs if r[0]))
        
        def pathmax(runs):
            return max(x[2] for x in runs if x[0])
        
        if lenmax is None:
            lenmax = self._index.getReadSize()
            
        brothers = sons(startnode)
        graph = self._graph
            
        runs = [[True,
                 x,
                 len(graph.getEdgeAttr(startnode,x)['label']),
                 [x]] 
                 for x in brothers]
                
        inter = intersect(runs)
        newpath=True

        while newpath and len(inter)==0 and pathmax(runs) < lenmax:
            newruns=[]
            newpath=False
            for ok,node,plenth,path in runs:
                if ok:
                    brothers = sons(node)
                    extraruns = [[True,x,
                                  plenth + len(graph.getEdgeAttr(node,x)['label']),
                                  path + [x]
                                 ]
                                 for x in brothers]

                    if extraruns: 
                        toappend=[x for x in extraruns if x]
                        if toappend:         
                            newruns.extend(toappend)
                            newpath=True
                    else:
                        newruns.append([False,node,plenth,path])
                else:
                    newruns.append([False,node,plenth,path])
                    
            runs = newruns
                
            if len(runs) > wide:
                return None,None
            
            if newpath:
                inter = intersect(runs)
                        
        if newpath:
            if len(inter)==0:
                return None,None
            
            pathok = [x[3] for x in runs if x[0]]
            
            out = set(y[min(y.index(z) for z in inter)] for y in pathok)
    
            assert len(out)==1,"Bug in the bubble detection during analysis of the bubblue starting in %d" % startnode
            
            out = out.pop()
            
            graph.getNodeAttr(out)['type']="microbubble out"
        
            for ok,node,plenth,path in runs:
                try:
                    del path[path.index(out):]
                except ValueError:
                    pass
                
            runmax=[]
            
            for path in pathok:
                cov = 0
                for s in path:
                    n = graph.getNodeAttr(s)
                    cov+=n['weight']
                runmax.append(cov)    
                
            
            goodpath = pathok[runmax.index(max(runmax))]
            
            
            b = startnode
            graph.getNodeAttr(b)['color']="green"
        
            for e in goodpath:
                graph.getNodeAttr(e)['color']="green"
                graph.getEdgeAttr(b,e)['color']='green'
                b=e
                    
            graph.getEdgeAttr(b,out)['color']='green'
            graph.getNodeAttr(out)['color']="green"
        
            nodeset = set([startnode,out]) 
        
            for r in runs:
                nodeset |= set(r[3])
        
            badset=(nodeset - set(goodpath)) - set([startnode,out])

        else:
            graph.getNodeAttr(startnode)['color']="green"
            nodeset = set([startnode]) 
        
            for r in runs:
                nodeset |= set(r[3])
        
            badset=nodeset - set([startnode])
            out=None
            
        
        for bad in badset:
            badnode = graph.getNodeAttr(bad)
            badnode['color']="red"
            badnode['type']="to remove"
    
    
        return self._graph.subgraph('microbubble',nodeset),out
        
    def microBubbleIterator(self,lenmax=None, wide=100):
        
        ibubble = self.forkIterator()
        
        for f in ibubble:
            b,out = self.microBubble(f,lenmax,wide)
            if b is not None:
                bnode = set(b) 
                if out is not None:
                    bnode.remove(out)
                ibubble|=bnode
                    
                yield b
                    
                            
    def resolveMicroBubble(self,lenmax=None, wide=100):
        
        n = 0
        
        print >>sys.stderr,"Resolving micro bubbles"
        
        for b in self.microBubbleIterator(lenmax):
            n+=1
            print >>sys.stderr,"Bubble : %8d\r" %n,
            
            if len(b) > 40:
                print >>open('bubble.dot','w'),b
                print >>sys.stderr,"Bubble : %8d >>> new graph %10s\r" % (n,b.nodeIterator().next()),
                
        print >>sys.stderr
        
        todel = [x for x in self._graph.nodeIterator(lambda n:"type" in self._graph[n] and self._graph[n]["type"]=="to remove")]

        print >>sys.stderr,"Cleanning %8d nodes" % len(todel)
        
        for n in todel:
            self._graph.deleteNode(n) 
            
    def endNodeSet(self):
        def isfinalnode(node):
            return len(set(self._graph.neighbourIterator(node,edgePredicate=lambda e : self._graph[e]['type']=="next read")))==0
    
        return set(x for x in self._graph.nodeIterator(isfinalnode))

    def cleanDeadBranches(self):
        def isfinalnode(node):
            return len(set(self._graph.neighbourIterator(node,edgePredicate=lambda e : self._graph[e]['type']=="next read")))==0
        
        print >>sys.stderr,"Cleanning dead branches"
        endpoints = self.endNodeSet()
        
        i=0
                
        for p in endpoints:
            tocut = [p]
            while tocut:
                nextcut=[]
                for p in tocut:
                    nextcut.extend(self._graph.parentIterator(p))
                    self._graph.deleteNode(p)
                    i+=1
                tocut = set(p for p in nextcut if isfinalnode(p))
                
                print >>sys.stderr,"%9d deleted nodes\r" % i,
                    
                
        print >>sys.stderr
                
        return i
    
    def simplifyAssembly(self,lenmax=None, wide=100):
                
        self.cleanDeadBranches()
        
        l = 0
        
        while (l!=len(self._graph)):
            l=len(self._graph)
            self.resolveMicroBubble(lenmax, wide)
            self.cleanDeadBranches()
            
    def __len__(self):
        return len(self._graph)
    
    def stemIterator(self):
        
        def sons(node):
            return set(self._graph.neighbourIterator(node,edgePredicate=lambda e : self._graph[e]['type']=="next read"))
            
        def parents(node):
            return set(self._graph.parentIterator(node,edgePredicate=lambda e : self._graph[e]['type']=="next read"))

        def is_junction(node):
            return  len(set(self._graph.neighbourIterator(node,edgePredicate=lambda e : self._graph[e]['type']=="next read"))) != 1 or \
                    len(set(self._graph.parentIterator(node,edgePredicate=lambda e : self._graph[e]['type']=="next read"))) != 1
        
        ni=self._graph.nodeIterator(is_junction)

        for first in ni:
            branches=sons(first)
            
            for n in branches:
                
                brothers=sons(n)
                fathers=parents(n)
                sequence=[]
                
                sequence.append(self._graph[(first,n)]['label'])
                                
                while len(brothers)==1 and len(fathers)==1:
                    son=brothers.pop()
                    sequence.append(self._graph[(n,son)]['label'])
                    n=son
                    brothers=sons(n)
                    fathers=parents(n)
    
                last = n
                sequence=''.join(sequence)
                lseq = len(sequence)
                
                yield first,last,lseq,sequence

        
    def compactAssembling(self):
        graph = DiGraph('compact')
        
        n = 0
        lcontig=0
        print >>sys.stderr,"Compacting graph :"
        
        for first,last,lseq,sequence in self.stemIterator():
            n+=1
            lcontig+=lseq
            print >>sys.stderr," Stem %9d : %6d bp (total : %6d)" % (n,lseq,lcontig)
            graph.addEdge(first,last)
            attr = graph.getEdgeAttr(first,last)
            attr['sequence']=sequence
            attr['length']=lseq
            if lseq > 10:
                attr['label']="%s->(%d)->%s" % (sequence[0:5],lseq,sequence[-5:])
            else:
                attr['label']="%s->(%d)" % (sequence,lseq)
                
        return graph
                       
              
def compactAssembling(self):

    def sons(node):
        return set(self._graph.neighbourIterator(node,edgePredicate=lambda e : self._graph[e]['type']=="next read"))
        
    def parents(node):
        return set(self._graph.parentIterator(node,edgePredicate=lambda e : self._graph[e]['type']=="next read"))
    
    graph = DiGraph('compact')

    first = self._graph.nodeIterator().next()
    brothers=sons(first)
    fathers=parents(first)
        
    while len(brothers)==1 and len(fathers)==1:
        first = fathers.pop()
        brothers=sons(first)
        fathers=parents(first)
        
    graph.addNode(first)
    
    brothers=sons(first)   
    ext = [(first,x,[self._graph[(first,x)]['label']]) for x in brothers]

    i=0
    while ext:
        i+=1
        nextext=[]
        print >>sys.stderr,"Extension points: %s" % ext
        for current,n,sequence in ext:
            brothers=sons(n)
            fathers=parents(n)
            
            while len(brothers)==1 and len(fathers)==1:
                son=brothers.pop()
                sequence.append(self._graph[(n,son)]['label'])
                n=son
                brothers=sons(n)
                fathers=parents(n)
                
            sequence=''.join(sequence)
            known = n in graph and graph.neighbourIterator(n)
            graph.addEdge(current,n)
            attr = graph.getEdgeAttr(current,n)
            attr['sequence']=sequence
            if len(sequence) > 10:
                attr['label']="%s->(%d)->%s" % (sequence[0:5],len(sequence),sequence[-5:])
            else:
                attr['label']="%s->(%d)" % (sequence,len(sequence))
            
            if not known:
                nextext.extend((son,x,[self._graph[(n,x)]['label']]) for x in brothers)
            print >>sys.stderr,"==>",nextext
        ext = nextext
        
    print >>sys.stderr
    return graph
                  
            
def endNodeSet(self):
    def isfinalnode(node):
        return len(set(self._graph.neighbourIterator(node,edgePredicate=lambda e : self._graph[e]['type']=="next read")))==0

    return set(x for x in self._graph.nodeIterator(isfinalnode))
        
def cleanDeadBranches(self):
    
    endpoint = self.endNodeSet()
    
 
def selectOneStepPath(self):

    def is_single(node):
        return  len(set(self._graph.neighbourIterator(node,edgePredicate=lambda e : self._graph[e]['type']=="next read"))) != 1 or \
                len(set(self._graph.parentIterator(node,edgePredicate=lambda e : self._graph[e]['type']=="next read"))) != 1
      