from stdio cimport snprintf
from stdlib cimport atol

from cpython.function cimport PyFunction_Check

from _graph cimport *

cdef bytes formatDict(dict keys):
    cdef bytes key
    cdef list output=[]
    
    for key,val in keys.iteritems():
        output.append(b'%s="%s"' % (key,bytes(val)))
        
    return b"[%s]" % b" ".join(output)


cdef class NodeIterator:

    def __init__(self,DiGraph graph, object predicate=None):
        
    #    if predicate is not None and not PyFunction_Check(predicate):
    #        raise TypeError,"predicate must be a function or None"
        
        self._nodepredicate = predicate
        self._graph = graph 
        self._exclude = set()
        self._nodeiterator = iter(graph._nodes)
        
    cpdef add(self,int node):
        cdef char[50] buffer
        cdef bytes key
        
        snprintf(buffer,50,"%d",node)
        key = buffer

        self._exclude.add(key)
        
    cpdef remove(self, int node):
        cdef char[50] buffer
        cdef bytes key
        
        snprintf(buffer,50,"%d",node)
        key = buffer

        self.exclude.remove(key)
        
                
    def __next__(self):
        cdef bytes key
        cdef int node
        cdef bint ok
        
        
        ok =False
        
        while(not ok):
            key = self._nodeiterator.next()
            node = atol(key)
            ok = (self._nodepredicate is None or self._nodepredicate(node)) \
                   and key not in self._exclude
                
        return node
    
    def __ior__(self,others):
        cdef int i
        
        for i in others:
            i=i+0
            self.add(i)
            
        return self

    def __iter__(self):
        return self
    
    

cdef class DiGraph:

    def __init__(self,bytes name):
        self._nodes={}
        self._edges_attrs={}
        self._nodes_attrs={}
        self._name=name 
        
    def __contains__(self,int node):
        cdef char[50] buffer
        cdef bytes key
        
        snprintf(buffer,50,"%d",node)
        key = buffer
        
        return key in self._nodes
    
    cpdef addNode(self,int node):
        cdef char[50] buffer
        cdef bytes key
        
        snprintf(buffer,50,"%d",node)
        key = buffer
        
        if key not in self._nodes:
            self._nodes[key]=(set(),set())
            self._nodes_attrs[key]={}
            
    cpdef addEdge(self,int node1, int node2):
        cdef char[50] buffer1
        cdef char[50] buffer2
        cdef bytes key1 
        cdef bytes key2 
        cdef tuple edge
        
        snprintf(buffer1,50,"%d",node1)
        key1 = buffer1
        
        snprintf(buffer2,50,"%d",node2)
        key2 = buffer2
        
        edge = (key1,key2)
        
        if edge not in self._edges_attrs:
            self.addNode(node1)
            self.addNode(node2)
        
            self._nodes[key1][0].add(key2)
            self._nodes[key2][1].add(key1)

            self._edges_attrs[edge]={}
            
    cpdef dict getEdgeAttr(self, int node1, int node2):
 
        cdef char[50] buffer1
        cdef char[50] buffer2
        cdef bytes key1 
        cdef bytes key2 
        cdef tuple edge
        cdef dict attr
        
        snprintf(buffer1,50,"%d",node1)
        key1 = buffer1
        
        snprintf(buffer2,50,"%d",node2)
        key2 = buffer2
   
        edge = (key1,key2)
        
        attr = self._edges_attrs[edge]
        
        return attr 
    
    cpdef dict getNodeAttr(self, int node):
        cdef char[50] buffer
        cdef bytes key
        cdef dict attr
        
        snprintf(buffer,50,"%d",node)
        key = buffer
        
        attr = self._nodes_attrs[key]
        
        return attr 

    cpdef int nodeCount(self):
        return len(self._nodes)
    
    cpdef int edgeCount(self):
        return len(self._edges_attrs)
    
    cpdef bytes dot(self,nodePredicate=None, edgePredicate=None):
        cdef bytes dot
        cdef int node1
        cdef int node2
        cdef list lines=[]
        cdef bytes line
        cdef dict attr 
        
        for node1 in self.nodeIterator(nodePredicate):
            attr = self.getNodeAttr(node1)
            if attr:
                line = b"   %10d %s" % (node1,formatDict(attr))
            else:
                line = b"   %10d" % node1 
            lines.append(line)
            
        for node1,node2 in self.edgeIterator(nodePredicate,edgePredicate):
            attr = self.getEdgeAttr(node1,node2)
            if attr:
                line = b"   %10d -> %10d %s" % (node1,node2,formatDict(attr))
            else:
                line = b"   %10d -> %10d" % (node1,node2) 
            lines.append(line)
        
        dot = b"\n".join(lines)
        dot = b"digraph %s {\n%s\n}" % (self._name,dot)
        
        return dot

    cpdef deleteEdge(self, int node1, int node2):
        cdef char[50] buffer1
        cdef char[50] buffer2
        cdef bytes key1 
        cdef bytes key2 
        cdef tuple edge

        
        snprintf(buffer1,50,"%d",node1)
        key1 = buffer1
        
        snprintf(buffer2,50,"%d",node2)
        key2 = buffer2
   
        edge = (key1,key2)
        
        del self._edges_attrs[edge]
        
        self._nodes[key1][0].remove(key2)
        self._nodes[key2][1].remove(key1)
        
    cpdef deleteNode(self, int node):
        cdef char[50] buffer
        cdef bytes key
        cdef dict attr
        cdef int node2
        cdef list edges
        
        snprintf(buffer,50,"%d",node)
        key = buffer
        
        del self._nodes_attrs[key]
        
        edges = [node2 for node2 in self.neighbourIterator(node)]
        for node2 in edges:
            self.deleteEdge(node,node2)
            
        edges = [node2 for node2 in self.parentIterator(node)]
        for node2 in edges:
            self.deleteEdge(node2,node)
            
        del self._nodes[key]
            
    cpdef DiGraph subgraph(self,bytes name, nodes):
        cdef DiGraph graph
        cdef set nodeset = set()
        cdef int node
        cdef int node2
        cdef char[50] buffer1
        cdef char[50] buffer2
        cdef bytes key1
        cdef bytes key2
        cdef tuple edge
        cdef set n0
        cdef set n1
        
        graph = self.__class__(name)
        
        for node in nodes:
            snprintf(buffer1,50,"%d",node)
            key1 = buffer1
            nodeset.add(key1)
            
        for key1 in nodeset:
            n0 = self._nodes[key1][0]
            n1 = self._nodes[key1][1]
            
            graph._nodes[key1] = (n0 & nodeset, n1 & nodeset)
            graph._nodes_attrs[key1] = self._nodes_attrs[key1]
            
            for key2 in graph._nodes[key1][0]:
                edge=(key1,key2)                
                graph._edges_attrs[edge]=self._edges_attrs[edge]
        
        return graph
    
    cpdef set connectedComponent(self,int node, nodePredicate=None, edgePredicate=None):
        cdef set component=set([node])
        cdef set toexplore=set([node])
        cdef int n
        cdef int v 
        
        assert nodePredicate is None or nodePredicate(node)
        
        
        while toexplore:
            n = toexplore.pop()
            for v in self.neighbourIterator(n,nodePredicate, edgePredicate):
                if v not in component:
                    toexplore.add(v)
                    component.add(v)

            for v in self.parentIterator(n,nodePredicate, edgePredicate):
                if v not in component:
                    toexplore.add(v)
                    component.add(v)
                    
        return component
                
        
        
    
    def nodeIterator(self,predicate=None):
        return NodeIterator(self,predicate)
                
    def edgeIterator(self,nodePredicate=None, edgePredicate=None):
        cdef tuple edge
        cdef bytes key1
        cdef bytes key2
        cdef int   node1
        cdef int   node2
        
        for edge in self._edges_attrs:
            key1 = edge[0]
            key2 = edge[1]
            
            node1 = atol(key1)
            node2 = atol(key2)
            
            if nodePredicate is None or (nodePredicate(node1) and nodePredicate(node2)):
                edge = (node1,node2)
                if edgePredicate is None or edgePredicate(edge):
                    yield edge
                    
    
            edge = (node1,node2)
            
    def parentIterator(self,int node,nodePredicate=None, edgePredicate=None):
        cdef char[50] buffer
        cdef bytes key
        cdef set neighbours
        cdef int n 
        
        
        snprintf(buffer,50,"%d",node)
        key = buffer
        
        neighbours = self._nodes[key][1]
        
        for key in neighbours:
            n = atol(key)
            if nodePredicate is None or nodePredicate(n):
                if edgePredicate is None or edgePredicate((n,node)):
                    yield n
                
    def neighbourIterator(self,int node,nodePredicate=None, edgePredicate=None):
        cdef char[50] buffer
        cdef bytes key
        cdef set neighbours
        cdef int n 
        
        
        snprintf(buffer,50,"%d",node)
        key = buffer
        
        neighbours = self._nodes[key][0]
        
        for key in neighbours:
            n = atol(key)
            if nodePredicate is None or nodePredicate(n):
                if edgePredicate is None or edgePredicate((node,n)):
                    yield n
                    

    def connectedComponentIterator(self,nodePredicate=None, edgePredicate=None):
        cdef NodeIterator ni=self.nodeIterator(nodePredicate)
        cdef set cc
        cdef int n
        
        for n in ni:
            cc = self.connectedComponent(n,nodePredicate,edgePredicate)
            ni|=cc
            yield cc
            
        
        
    def __len__(self):
        return self.nodeCount()
    
    def __iter__(self):
        return self.nodeIterator()
            
    def __getitem__(self,key):
        cdef int   node1
        cdef int   node2
                
        if isinstance(key,tuple):
            node1 = key[0]
            node2 = key[1]
            
            return self.getEdgeAttr(node1,node2)
        
        elif isinstance(key,int):
            node1 = key

            return self.getNodeAttr(node1)
        
        raise TypeError,"key must be an int or a tuple"

    def __str__(self):
        return self.dot()
    
    def __delitem__(self, int node):
        self.deleteNode(node)

    __bytes__=__str__

        
        

        