from _orgasm cimport *

cdef extern from "stdlib.h":      
    ctypedef unsigned int size_t  
    int32_t abs(int32_t i)

cdef extern from "orgasm.h":
    void freeBuffer(buffer_t *buffer)
    
    buffer_t *loadIndexedReads(char *indexname)

    int32_t lookForString(buffer_t *buffer, char *key, size_t length, int32_t* endoflist)
    int32_t nextRead(buffer_t *buffer, int32_t current, size_t length, int32_t* endoflist)
    char *getRead(buffer_t *buffer, int32_t recordid, uint32_t begin, int32_t length, char *dest)

cdef class Index:

    def __init__(self,bytes name):
        self._index=loadIndexedReads(name)
        
    cpdef bint contains(self, bytes word):
        cdef int32_t endoflist
        cdef size_t  lword = len(word)
        
        lookForString(self._index,word,lword,&endoflist)
        
        return endoflist==0
                   
    cpdef int count(self,word):
        cdef size_t n=0
        cdef int32_t readid
        
        for readid in self[word]:
            n+=1
            
        return n
    
    cpdef int len(self):
        return self._index.readCount

    cpdef bytes getRead(self, int32_t readid, int32_t begin, int32_t length):
        cdef char buffer[500]
        cdef bytes seq
        
        assert begin < self._index.readSize,"Begin must be smaller than read size"
        assert begin+length <= self._index.readSize,"Begin+Length must be smaller than read size"
        
        if readid==0 or abs(readid)>=self._index.readCount:
            raise IndexError,readid
        
        seq=getRead(self._index, readid, begin, length, buffer)

        return seq

    cpdef dict getExtensions(self, bytes word):
        cdef dict rep
        cdef int32_t readid
        cdef size_t  lword = len(word)
        cdef bytes e
        cdef list  data
        
        rep = {}

        for readid in self[word]:
            e = self.getRead(readid,lword,self._index.readSize - lword)
            data = rep.get(e,[0,[]])
            data[0]+=1
            data[1].append(readid)
            rep[e]=data
        
        return rep

    cpdef int getReadSize(self):
        return self._index.readSize
    
    cpdef tuple getReadIds(self, bytes seq):
    
        cdef list x 
        cdef int  y
        cdef int  t
        
        assert len(seq)==self._index.readSize,"seq must have the same length than reads"
        
        x = [y for y in self[seq]]
        
        t = len(self)+1
        
        for y in x:
            if abs(y) < abs(t):
                t = y
                
        return (t,len(x),set(x))
    
    def __dealloc__(self):
        freeBuffer(self._index)
        
    def __contains__(self, bytes word):
        return self.contains(word)
    
    def __getitem__(self, bytes word):
        cdef int32_t endoflist
        cdef int32_t readid
        cdef size_t  lword = len(word)

        readid = lookForString(self._index,word,lword,&endoflist)

        while(endoflist==0):
            yield readid
            readid = nextRead(self._index,readid,lword,&endoflist)
            
    def __len__(self):
        return self.len()
                
   
      
        
        
        
    
        

